@@ -21,7 +21,8 @@
 (******************************************************************************)
 
 
-open Unix;;
+open Unix
+
 
 type t =
     {
@@ -32,7 +33,7 @@ type t =
       verbose:        string -> unit;
       mutable prev:   string;
     }
-;;
+
 
 type expect_match =
     [
@@ -44,11 +45,18 @@ type expect_match =
       | `Contains of string
       | `Timeout
     ]
-;;
+
 
 let default_verbose_output =
   print_endline
 
+
+let safe_close report fd =
+  try
+    close fd
+  with e ->
+    report (Printexc.to_string e)
+
 let spawn ?(verbose=false)
           ?(verbose_output=default_verbose_output)
           ?(timeout=Some 10.0)
@@ -64,63 +72,60 @@ let spawn ?(verbose=false)
          | 0 -> prg
          | n -> args.(n - 1))
   in
-  let proc_stdin, self_stdin =
-    pipe ()
+  let verbose str =
+    if verbose then
+      verbose_output str
   in
-  let self_stdout, proc_stdout =
-    pipe ()
+  let proc_stdin, self_stdin = pipe () in
+  let self_stdout, proc_stdout = pipe () in
+  let safe_set_close_on_exec fd = try set_close_on_exec fd with _ -> () in
+  let safe_close =
+    safe_close
+      (fun exn_str ->
+        verbose
+          (Printf.sprintf
+             "Error occurs when starting %s: %s" command_line exn_str))
   in
-  let safe_set_close_on_exec fd =
-    try
-      set_close_on_exec fd
-    with _ ->
-      ()
-  in
-  let proc_stderr =
-    if use_stderr then
-      begin
-        let fd =
-          dup proc_stdout
-        in
-          safe_set_close_on_exec fd;
-          fd
-      end
-
-    else
-      begin
-        stderr
-      end
+  let proc_stderr, safe_close_stderr =
+    if use_stderr then begin
+      let fd = dup proc_stdout in
+      safe_set_close_on_exec fd;
+      fd, fun () -> safe_close fd
+    end else begin
+      stderr, ignore
+    end
   in
   let pid =
     safe_set_close_on_exec self_stdin;
     safe_set_close_on_exec self_stdout;
+    verbose (Printf.sprintf "Starting command line: %s" command_line);
     match env with
       | Some a ->
           create_process_env prg cmd a proc_stdin proc_stdout proc_stderr
       | None ->
           create_process prg cmd proc_stdin proc_stdout proc_stderr
   in
-  let verbose str =
-    if verbose then
-      verbose_output str
-  in
-    close proc_stdout;
-    close proc_stdin;
-    verbose ("Command line: "^command_line);
-
+  let t =
+    safe_close proc_stdout;
+    safe_close proc_stdin;
+    safe_close_stderr ();
     {
       pid           = pid;
       expect_stdin  = self_stdin;
       expect_stdout = self_stdout;
       timeout       = timeout;
-      verbose       = verbose;
+      verbose       = (fun str ->
+                         verbose (Printf.sprintf "PID(%d): %s" pid str));
       prev          = "";
     }
-;;
+  in
+    t.verbose "Started";
+    t
+
 
 let set_timeout t timeout =
   {t with timeout = timeout}
-;;
+
 
 let send t str =
   let _i : int =
@@ -128,20 +133,20 @@ let send t str =
     write t.expect_stdin str 0 (String.length str)
   in
     ()
-;;
+
 
 type expect_event =
   | Eof
   | Timeout
   | Line of string
-;;
+
 
 let expect t ?(fmatches=[]) actions action_default =
   let buff =
     String.make 4096 'x'
   in
 
-  (* Tets if an event can be associated with a fmatch action or continue *)
+  (* Test if an event can be associated with a fmatch action or continue *)
   let action_fmatch event cont =
     match event with
       | Eof | Timeout ->
@@ -206,59 +211,44 @@ let expect t ?(fmatches=[]) actions action_default =
       try
         read t.expect_stdout buff 0 (String.length buff)
       with End_of_file
-        | Unix_error(EPIPE, "read", _) ->
+        | Unix_error(EPIPE, "read", _)
+        | Sys_error("Broken pipe") ->
         0
     in
 
-    let input_str =
-      String.sub buff 0 input_len
-    in
-
-    let () =
+    if input_len = 0 then begin
+      (* Nothing to process anymore *)
+      t.verbose "Receive end of file.";
+      action_match Eof (fun () -> action_default)
+    end else begin
+      let input_str = String.sub buff 0 input_len in
       t.verbose (Printf.sprintf "Receive: %S" input_str);
-    in
+      let lines = BatString.nsplit (t.prev ^ input_str) "\n" in
+      let rec scan_lines =
+        function
+          | [ln] when
+              ln <> "" &&
+              ln.[(String.length ln) - 1] <> '\n' ->
+              begin
+                action_match (Line ln)
+                  (fun () ->
+                     t.prev <- ln;
+                     cont ())
+              end
 
-    (* Modify continuation if we reach the end of the stream *)
-    let cont =
-      if input_len = 0 then
-        (* Nothing to process anymore *)
-        fun () ->
-          action_match
-            Eof
-            (fun () -> action_default)
-      else
-        cont
-    in
-
-    let lines =
-      BatString.nsplit (t.prev ^ input_str) "\n"
-    in
-
-    let rec scan_lines =
-      function
-        | [ln] when
-            ln <> "" &&
-            ln.[(String.length ln) - 1] <> '\n' ->
-            begin
-              action_match (Line ln)
-                (fun () ->
-                   t.prev <- ln;
-                   cont ())
-            end
-
-        | hd :: tl ->
-            begin
-              action_match (Line hd) (fun () -> scan_lines tl)
-            end
-
-        | [] ->
-            begin
-              t.prev <- "";
-              cont ()
-            end
-    in
-      scan_lines lines
+          | hd :: tl ->
+              begin
+                action_match (Line hd) (fun () -> scan_lines tl)
+              end
 
+          | [] ->
+              begin
+                t.prev <- "";
+                cont ()
+              end
+      in
+        scan_lines lines
+    end
   in
 
   (* Read lines from process, considering timeout *)
@@ -266,23 +256,19 @@ let expect t ?(fmatches=[]) actions action_default =
     let time_left =
       timeout -. ((Unix.gettimeofday ()) -. time_begin)
     in
-      if time_left > 0.0 then
-        begin
-          match Unix.select [t.expect_stdout] [] [] time_left with
-            | [], _, _ ->
-                expect_input_line_timeout
-                  timeout time_begin
+      if time_left > 0.0 then begin
+        match Unix.select [t.expect_stdout] [] [] time_left with
+          | [], _, _ ->
+              expect_input_line_timeout
+                timeout time_begin
 
-            | _ ->
-                expect_input_line
-                  (fun () ->
-                     expect_input_line_timeout
-                       timeout time_begin)
-        end
-      else
-        begin
-          action_match Timeout (fun () -> action_default)
-        end
+          | _ ->
+              expect_input_line
+                (fun () -> expect_input_line_timeout timeout time_begin)
+      end else begin
+        t.verbose "Receive timeout";
+        action_match Timeout (fun () -> action_default)
+      end
   in
 
   (* Read lines from process *)
@@ -294,12 +280,10 @@ let expect t ?(fmatches=[]) actions action_default =
     (* Read a stdout until something is found *)
     match t.timeout with
       | Some timeout ->
-          expect_input_line_timeout
-            timeout
-            (Unix.gettimeofday ())
+          expect_input_line_timeout timeout (Unix.gettimeofday ())
       | None ->
           expect_input_line_notimeout ()
-;;
+
 
 let close t =
   let rec waitpid_non_intr () =
@@ -308,16 +292,26 @@ let close t =
     with Unix_error (EINTR, _, _) ->
       waitpid_non_intr ()
   in
-  let safe_close fd =
-    try
-      close fd
-    with _ ->
-      ()
-  in
-    safe_close t.expect_stdout;
-    safe_close t.expect_stdin;
+  let status =
+    safe_close
+      (fun exn_str ->
+         t.verbose (Printf.sprintf "Fail to close stdout: %s" exn_str))
+      t.expect_stdout;
+    safe_close
+      (fun exn_str ->
+         t.verbose (Printf.sprintf "Fail to close stdin: %s" exn_str))
+      t.expect_stdin;
     snd (waitpid_non_intr ())
-;;
+  in
+    t.verbose
+      (Printf.sprintf
+         "Stopping: %s"
+         (match status with
+            | WEXITED i -> Printf.sprintf "WEXITED %d" i
+            | WSIGNALED i -> Printf.sprintf "WSIGNALED %d" i
+            | WSTOPPED i -> Printf.sprintf "WSTOPPED %d" i));
+    status
+
 
 let with_spawn
       ?verbose ?verbose_output ?timeout ?env ?use_stderr prog args f a =
@@ -325,16 +319,10 @@ let with_spawn
     spawn ?verbose ?verbose_output ?timeout ?env ?use_stderr prog args
   in
     try
-      let res =
-        f t a
-      in
-      let exit_code =
-        close t
-      in
-        res, exit_code
-
+      let res = f t a in
+      let exit_code = close t in
+      res, exit_code
     with e ->
       ignore(close t);
       raise e
-;;
 
